# `go build`命令

`go build`命令顾名思义就是用来对项目进行编译的，它的命令格式为`go build [-o 输出目录] [编译参数] [待编译的包]`。

## 更改输出目录

默认情况下，`go build`会将编译输出的结果文件放置在与待编译包或者文件相同的目录中，而且名称也相同。例如在 Windows 系统中执行`go build main.go`来对`main.go`文件进行编译，那么编译后的结果文件就会与`main.go`文件在一起，且名称为`main.exe`。

如果需要更改`go build`的默认输出目录，就需要使用`-o`参数，`-o`参数中一般只需要指定输出目录即可，编译输出的文件名会与输出的文件名相同。

这里在输出目录和输出文件名的指定上有以下需要注意的情况：

1. 编译命令为`go build -o ./build main.go`，但`./build`目录不存在，那么此时`main.go`文件将会被编译为一个名为`build`的可执行文件。
1. 编译命令为`go build -o ./build main.go`，但`./build`目录已经存在，那么此时`main.go`文件将会被编译为一个保存在`./build`目录下的名为`main`的可执行文件。
1. 编译命令为`go build -o ./build/ main.go`，但`./build`目录不存在，那么此时`go build`命令将会创建`build`目录，然后将编译好的名为`main`的可执行文件放置在其中。
1. 编译命令为`go build -o ./build/ main.go`，但`./build`目录已经存在，那么此时`go build`命令将编译好的名为`main`的可执行文件放置在指定的`./build`目录中。

以上四种情况就是在指定编译目标的时候，带与不带结尾的`/`的区别，在实际编译项目的时候需要注意。

## 其他常用编译参数

一般情况下直接使用`go build`对项目进行编译就已经能够满足要求了，但是在一些特殊情况下，对于项目的编译还需要进行一些自定义，这是就需要使用到额外的编译参数了。

- `-a`，强制重新编译所有的包。
- `-n`，输出要执行的命令但并不实际执行。
- `-race`，使用数据竞争检测。该选项有平台限制，使用的时候需要查看所使用 Go 版本的说明。
- `-msan`，启用与内存清理器的交互。
- `-asan`，启动与地址清理器的交互。
- `-work`，输出临时工作目录并予以保留。
- `-buildmode mode`，设定要使用的编译模式。
- `-compiler name`，舍弟索要使用的编译器的名称，可选`gccgo`或者`gc`。
- `gccgoflags '[pattern=]arg list'`，设定使用`gccgo`编译器的时候需要传递给编译器的参数。
- `gcflags '[pattern=]arg list'`，设定使用`gc`编译器的时候需要传递给编译器的参数。
- `-ldflags [pattern=]arg list'`，设定每次工具链接调用时传递的参数。
- `-linkshared`，设定在构建前链接之前使用`-buildmode=shared`构建的动态链接代码。
- `-mod mode`，设定模块的下载使用模式。
- `-tags tag,list`，使用一个逗号分割的列表来设定用于约束编译的条件。

## 编译输出模式

在前面一节中，常用的参数提到了可以通过`-buildmode`来设定编译器的编译输出模式，常用的编译输出模式主要有 C 静态链接库、静态链接库、C 动态链接库、动态链接库、可执行文件等。

### C 静态链接库

输出 C 静态链接库需要使用`-buildmode=c-archive`参数。

C 静态链接库可以使用`//export 导出函数名`标记来标记`main`包中的函数，并将其作为静态链接库的导出函数供其他 C 程序使用。

例如编写一个简单的程序。

```go
package main

import (
    "fmt"
    "C"
)

func main() {}

//export Add
func Add(a, b int) int {
    return a+b
}
```

这个示例文件`add.go`在编译为 C 静态链接库以后，将会生成`add.h`和`add.a`两个文件。熟悉 C 语言的朋友可以直接使用 C 语言中调用静态链接库的方法来使用生成的静态链接库。

### C 动态链接库

输出 C 动态链接库需要使用`-buildmode=c-shared`参数。

用来生成动态链接库的源代码文件与之前用来生成静态链接库的源代码文件是一致的，而且其中的标记也是一致的。但是编译后产生的文件则分别是`add.so`（Windows 平台上是`add.dll`）和`add.h`。

### Go 动态链接库

输出 Go 动态链接库需要使用`-buildmode=shared`参数。这样构建出来的讲师仅能够由 Go 使用的动态链接库。

例如可以执行以下命令将 Go 标准库编译安装为动态链接库，`go install -buidlmode=shared std`，之后就可以使用`-linkshred`编译我们的程序，例如以下这个。

```go
package main

import "fmt"

func main() {
    fmt.Println("Hello world")
}
```

此时，使用命令`go build -linkshared main.go`进行编译，可以发现生成的可执行文件体积比不使用`-linkshared`要少非常多，而使用`ldd`命令对其调用链介于分析，也可以看到它依赖了许多动态链接库文件。

但是这种方式生成的可执行文件体积虽然小，但是缺点也非常的多，如果系统中缺少了程序所依赖的动态链接库，或者动态链接库文件版本不匹配，那么程序都不会正常运行。

### 插件模式

插件模式是 Go 1.8 推出的一个新的构建方式，这个模式使用`-buildmode=plugin`进行编译。插件模式可以将一个`package main`包编译为一个动态链接库文件，然后可以供其他程序使用标准库中的`plugin`包功能在运行时动态加载。

例如有以下一个文件，其中提供了一个可以供其他程序调用的函数。

```go
package main

import "fmt"

type greeting string

func (g greeting) Say(s string) {
    fmt.Printf("Hello %s.\n", s)
}

var Greeter greeting
```

现在使用命令将其编译为一个插件，执行`go build -buildmode=plugin -o greeter.so greeter.go`。

接下来就可以在其他的程序中使用标准库中的`plugin`包中的功能进行动态加载了。

```go
package main

import (
    "fmt"
    "os"
    "plugin"
)

type Greet interface {
    Say(string)
}

func main() {
    plug, err := plugin.Open("./greeter.so")
    if err != nil {
        fmt.Println(err)
        os.Exit(1)
    }
    symObj, err := plug.Lookup("Greeter")
    if err != nil {
        fmt.Println(err)
        os.Exit(1)
    }
    var greeter Greet
    greeter, ok := symObj.(Greet)
    if !ok {
        fmt.Println(err)
        os.Exit(1)
    }
    greeter.Say("world")
}
```

```admonish caution
插件模式目前仅支持Linux、FreeBSD和macOS系统，Windows系统中是不支持的。
```

## 条件编译

Go 中对于条件编译的支持，是通过一条特殊的注释来实现的，这个注释以`//go:build`开头，位于`package`语句之前，它可以控制整个文件是否会被包含在编译过程中。

例如以下注释：

```go
//go:build (linux && 386) || (darwin && !cgo)
```

这个注释就约束了这个文件只会在`go build`命令的`-tags`参数中，包含`"linux,386"`或者`"darwin"`且不包含`"cgo"`标签的时候才会进行编译，例如执行编译命令`go build -tags="linux,386"`的时候，这个文件就会参与到编译中，而执行命令`go build -tags="win,x64"`时，这个文件就会被略过。

条件编译的标签组合操作符与普通的逻辑表达式的操作符是一样的，`&&`表示并且，`||`表示或者，`!`表示非。`//go:build`注释在一个文件中可以出现多行，多个约束条件之间默认采用`&&`的关系。如果编译条件被设定为`//go:build ignore`，那么这个文件将不会被包含在任何编译中。

如果在`//go:build`中使用的是操作系统和平台架构相关的限制内容，那么编译约束除了可以在`-tags`参数中提供，还将受到`GOOS`和`GOARCH`两个环境变量的影响。

如果一个源代码文件中的内容只会被用在某一个操作系统或者某一个平台架构中，那么还有一种更加简单的办法来对其进行约束。在 Go 中，源代码文件的文件名在去除`.go`的后缀以后，其结尾使用下划线分隔开的部分内容对于 Go 语言是有意义的，所以在 Go 源代码文件的命名中，不要轻易的使用下划线。这里对源代码文件的命名规则做一个简单的介绍。

- 文件以`_test`结尾，则表明这是一个用于单元测试的文件，会在执行`go test`命令时被执行。
- 文件以`_GOOS`系列内容结尾，例如`_windows`、`_darwin`、`_linux`则表示该文件需要在`GOOS`环境变量为相应值的时候参与编译。
- 文件以`_GOARCH`系列内容结尾的时候，例如`_386`、`_amd64`，则表示该文件需要在`GOARCH`环境变量为对应值的时候参与编译。
- 文件以`_GOOS_GOARCH`格式内容结尾时，其中将同时匹配`GOOS`和`GOARCH`两个环境变量中的内容，例如`_windows_amd64`。

```admonish tip
`GOARCH`环境变量还可以定义其下的架构特征作为更加详细的过滤条件，例如`GOARCH=386, GO386=sse2`，这就会将`386.sse2`也列入到编译条件中。具体可以使用的结构特征可以参考Go官方文档中的说明。
```

## 交叉编译

交叉编译又称为跨平台编译，是一种可以在一个平台上生成另一个平台使用的可执行程序的功能。Go 的工具链原生支持交叉编译功能。

使用 Go 的交叉编译十分简单，只需要在编译的时候临时改变以下环境变量的值即可。与交叉编译相关的环境变量主要有以下三个。

- `CGO_ENABLED`，用于指示 Go 编译器是否启用 CGO 功能，对于使用交叉编译功能来说，CGO 功能是不可用的，所以要恒定设置为`CGO_ENABLED=0`。
- `GOOS`，用于指示 Go 编译器，所编译的目标操作系统。
- `GOARCH`，用于指示 Go 编译器，所编译的目标平台架构。

所以在 macOS 上编译 Windows 和 Linux 系统的可执行程序就可以使用以下两个命令。

```bash
# 编译Windows平台程序
CGO_ENABLED=0 GOOS=windows GOARCH=amd64 go build main.go
# 编译Linux平台程序
CGO_ENABLED=0 GOOS=linux GOARCH=amd64 go build main.go
```

在 Linux 上编译 Windows 和 macOS 上的可执行程序可以使用以下两个命令。

```bash
# 编译Windows平台程序
CGO_ENABLED=0 GOOS=windows GOARCH=amd64 go build main.go
# 编译macOS平台程序
CGO_ENABLED=0 GOOS=darwin GOARCH=arm64 go build main.go
```

在 Windows 上编译 Linux 和 macOS 上的可执行程序就有些不一样，但是核心原理还是一样的。

```bash
# 编译Linux平台程序
SET CGO_ENABLED=0
SET GOOS=linux
SET GOARCH=amd64
go build main.go
# 编译macOS平台程序
SET CGO_ENABLED=0
SET GOOS=darwin
SET GOARCH=arm64
go build main.go
```

因为交叉编译每次都需要设置环境变量，所以建议在需要使用交叉编译的时候，可以将编译指令编写成一个脚本文件来简化编译过程。
